% Przyklad 6.11
function przyklad_06_11()

K = 3;                                                 % Liczba stopni w strukturze.
rand('state', sum(100 * clock));          % Inicjalizacja generatora liczb losowych.
alpha_start = rand(1, (K - 1) * 2) .* 2 .* pi; % Losowanie rozwiazania poczatkowego.

% Wielokrotne wywolanie algorytmu optymalizacyjnego z roznymi ...
% rozwiazaniami poczatkowymi zwieksza szanse na znalezienie globalnego optimum.

options = optimset('MaxFunEvals', 2000, 'DiffMinChange', 0.0001, 'TolFun',1e-19, ...
    'DiffMaxChange', 0.75 * pi, 'TolX', 1e-9);  % Parametry algorytmu optymalizacji.
alpha_opt = fminunc(@obj_fun, alpha_start, options); % Optymalizacja bez ograniczen.
obj_fun(alpha_opt); %              Prezentacja najlepszego znalezionego rozwiazania.

%===================================================================================
function err = obj_fun(a)                              % Funkcja celu optymalizacji.

Sigma = [eye(2)  eye(2)
         eye(2) -eye(2)];            % Macierz reprezentujaca obliczenia "motylkowe".

% Generowanie odpowiedzi impulsowych filtrow okreslonych zadanym wektorem
% parametrow.

h = dctIImtx(4);                        % Stopien podstawowy struktury: DCT typu II.
h = h([1 3 2 4], :); % Zmiana uporzadkowania wyjsc DCT: parzyste przed nieparzystymi.
h = diag([1 -1 -1 1]) * h;  % Zmiany znakow czasami poprawiaja wyniki optymalizacji.
for k = 1 : (length(a) / 2);             % Rozszerzenie struktury o kolejne stopnie.
    h = Sigma * h / 2;            % Obliczenia "motylkowe" polaczone z normalizacja.
    h = [h(1 : 2, :) zeros(2, 4)    % Opoznienie polowy odpowiedzi impulsowych o ...
        zeros(2, 4) h(3 : 4, :)];       % 4 probki. Na rys. 6.32 opoznienie o 1 ...
            % probke, bo tozsamosc kaskadowa uzyta do przemieszczenia decymatorow.
    % Tutaj od razu generowane odpowiedzi impulsowe zamiast skladowych polifazowych.
    h = Sigma * h;                                         % Obliczenia "motylkowe".
    h(1 : 2, :) = rotmtx(a(2 * k - 1)) * h(1 : 2, :);  % Obrot opisany macierza U_k.
    h(3 : 4, :) = rotmtx(a(2 * k)) * h(3 : 4, :);      % Obrot opisany macierza V_k.
end                                % Przywrocenie uporzadkowania wyjsc: parzyste ...
h = h([1 3 2 4], :) ./ 2;                             % na przemian z nieparzystymi.

% Ocena filtrow: na ile charakterystyki amplitudowe odbiegaja od idealnych.

L = 512;                              % Liczba punktow charakterystyki amplitudowej.
H = fft(h, 2 * L, 2);             % Obliczenie charakterystyk amplitudowych filtrow.
Habs = abs(H(:, 1 : L));
Href = [ ones(1, L / 4) zeros(1, 3 * L / 4)                           % Pozadane ...
    zeros(1, L / 4) ones(1, L / 4) zeros(1, L / 2)        % charakterystyki ...
    zeros(1, L / 2) ones(1, L / 4) zeros(1, L / 4)               % amplitudowe.
    zeros(1, 3 * L / 4) ones(1, L / 4) ];
weight0 = sin(((0 : (L / 4 - 1)) + 0.5) *  pi / 2 / (L / 8));        % Ustalenie ...
weight1 = sin(((0 : (1 * L / 4 - 1)) + 0.5) *  pi / 2 / (1 * L / 4));   % wag bledu.
weight2 = sin(((0 : (2 * L / 4 - 1)) + 0.5) *  pi / 2 / (2 * L / 4));
weight3 = sin(((0 : (3 * L / 4 - 1)) + 0.5) *  pi / 2 / (3 * L / 4));
weight = [fliplr(weight1) weight3;          fliplr(weight1) weight0 weight2
    fliplr(weight2) weight0 weight1;    fliplr(weight3) weight1  ] .^ 0.5;
err = (Habs - Href) .* weight;
err = sum(err(:) .^ 2);       % Obliczenie wypadkowego bledu dla wszystkich kanalow.

if (nargout == 0)     % Jezeli funkcja wywolywana spoza algorytmu optymalizacji, ...
    fb_show_response(h, 512);               % to prezentacja charakterystyk filtrow.
    figure;     plot(weight')
end

%===================================================================================
function R = rotmtx(a)                                % Macierz obrotu o zadany kat.

c = cos(a);     s = sin(a);     R = [c -s; s c];